
import typing as t

if t.TYPE_CHECKING:
  from nr.parsing.core.scanner import Scanner

T = t.TypeVar('T')
U = t.TypeVar('U')


class TokenExtractor(t.Generic[T]):
  """ Interface for token extraction. """

  def get_token(self, scanner: 'Scanner') -> t.Optional[T]:
    """ Extract a token value from the current position of the scanner. """

    raise NotImplementedError(f'{type(self).__name__}.get_token() is not implemented')

  @staticmethod
  def of(impl: t.Callable[['Scanner'], t.Optional[T]]) -> 'TokenExtractor[T]':
    return _LambdaTokenExtractor(impl)

  def map(self, func: t.Callable[[T], U]) -> 'TokenExtractor[U]':
    return _MappedTokenExtractor(func, self)


class _LambdaTokenExtractor(TokenExtractor[T]):

  def __init__(self, impl: t.Callable[['Scanner'], t.Optional[T]]) -> None:
    self._impl = impl

  def __repr__(self) -> str:
    return f'{type(self).__name__}({self._impl})'

  def get_token(self, scanner: 'Scanner') -> t.Optional[T]:
    return self._impl(scanner)


class _MappedTokenExtractor(TokenExtractor[U]):

  def __init__(self, func: t.Callable[[T], U], inner: TokenExtractor[T]) -> None:
    self._func = func
    self._inner = inner

  def __repr__(self) -> str:
    return f'{type(self).__name__}({self._func}, {self._inner})'

  def get_token(self, scanner: 'Scanner') -> t.Optional[U]:
    value = self._inner.get_token(scanner)
    if value is not None:
      return self._func(value)
    return None




import enum
import re
import typing as t

T = t.TypeVar('T')
U = t.TypeVar('U')


class Cursor(t.NamedTuple):
  offset: int
  line: int
  column: int


class Seek(enum.Enum):
  SET = enum.auto()
  CUR = enum.auto()
  END = enum.auto()


class Scanner:
  """
  A utility class to scan through text.
  """

  def __init__(self, text: str) -> None:
    if not isinstance(text, str):
      raise TypeError(f'expected str, got {type(text).__name__}')
    self.text = text
    self.index = 0
    self.lineno = 1
    self.colno = 0

  def __repr__(self) -> str:
    return f'<Scanner at {self.lineno}:{self.colno}>'

  def __bool__(self) -> bool:
    return self.index < len(self.text)

  @property
  def pos(self) -> Cursor:
    return Cursor(self.index, self.lineno, self.colno)

  @pos.setter
  def pos(self, cursor: Cursor) -> None:
    """ Moves the scanner back (or forward) to the specified cursor location. """

    if not isinstance(cursor, Cursor):
      raise TypeError(f'expected Cursor object {type(cursor).__name__}')
    self.index, self.lineno, self.colno = cursor

  @property
  def char(self) -> str:
    """ Returns the current character. Returns an empty string at the end of the text. """

    if self.index >= 0 and self.index < len(self.text):
      return self.text[self.index]
    else:
      return type(self.text)()

  def seek(self, offset: int, mode: t.Union[str, Seek] = Seek.SET) -> None:  # NOSONAR
    """
    Moves the cursor of the Scanner to or by *offset* depending on the *mode*. The method is
    similar to a file's `seek()` method, but ensures that the line and column counts are tracked
    correctly.
    """

    if isinstance(mode, str):
      mode = Seek[mode.upper()]

    if mode not in Seek:
      raise ValueError(f'invalid mode: {mode!r}')

    # Translate the other modes into the 'set' mode.
    if mode == Seek.END:
      offset = len(self.text) + offset
      mode = Seek.SET
    elif mode == Seek.CUR:
      offset = self.index + offset
      mode = Seek.SET
    assert mode == Seek.SET

    if offset < 0:
      offset = 0
    elif offset > len(self.text):
      offset = len(self.text) + 1

    if self.index == offset:
      return

    # Update line/column counts. Figure which path is shorter:
    # 1) Start counting from the beginning of the file,
    if offset <= abs(self.index - offset):
      text, index, lineno, colno = self.text, 0, 1, 0
      while index != offset:
        # Find the next newline in the string.
        nli = text.find('\n', index)
        if nli >= offset or nli < 0:
          colno = offset - index
          index = offset
          break
        else:
          colno = 0
          lineno += 1
          index = nli + 1

    # 2) or step from the current position of the cursor.
    else:
      text, index, lineno, colno = self.text, self.index, self.lineno, self.colno

      if offset < index:  # backwards
        while index != offset:
          nli = text.rfind('\n', 0, index)
          if nli < 0 or nli <= offset:
            if text[offset] == '\n':
              assert (offset - nli) == 0, (offset, nli)
              nli = text.rfind('\n', 0, index-1)
              lineno -= 1
            colno = offset - nli - 1
            index = offset
            break
          else:
            lineno -= 1
            index = nli - 1
      else:  # forwards
        while index != offset:
          nli = text.find('\n', index)
          if nli < 0 or nli >= offset:
            colno = offset - index
            index = offset
          else:
            lineno += 1
            index = nli + 1

    assert lineno >= 1
    assert colno >= 0
    assert index == offset
    self.index, self.lineno, self.colno = index, lineno, colno

  def next(self) -> str:
    """ Move on to the next character in the text. """

    char = self.char
    if char == '\n':
      self.lineno += 1
      self.colno = 0
    else:
      self.colno += 1
    self.index += 1
    return self.char

  def readline(self) -> str:
    """ Reads a full line from the scanner and returns it. """

    start = end = self.index
    while end < len(self.text):
      if self.text[end] == '\n':
        end += 1
        break
      end += 1
    result = self.text[start:end]
    self.index = end
    if result.endswith('\n'):
      self.colno = 0
      self.lineno += 1
    else:
      self.colno += end - start
    return result

  def match(self, regex: t.Union[str, re.Pattern], flags: int = 0, *,
      _search: bool = False) -> t.Optional[t.Match[str]]:
    """
    Matches the *regex* from the current character of the *scanner* and returns the result. The
    scanners column and line numbers are updated respectively.
    """

    if isinstance(regex, str):
      regex = re.compile(regex, flags)
    match = (regex.search if _search else regex.match)(self.text, self.index)
    if not match:
      return None
    start, end = match.start(), match.end()
    if not _search:
      assert start == self.index
    else:
      start = self.index
    lines = self.text.count('\n', start, end)
    self.index = end
    if lines:
      self.colno = end - self.text.rfind('\n', start, end) - 1
      self.lineno += lines
    else:
      self.colno += end - start
    return match

  def search(self, regex: t.Union[str, re.Pattern], flags: int = 0) -> t.Optional[re.Match]:
    """
    Performs a regex search from the current position of the scanner. Note that searching in the
    scanner will potentially have you skip characters without consuming them.
    """

    return self.match(regex, flags, _search=True)

  def getmatch(self, regex: t.Union[str, re.Pattern], group: t.Union[int,str] = 0,
      flags: int = 0) -> t.Optional[str]:
    """
    The same as #Scanner.match(), but returns the captured group rather than
    the regex match object, or None if the pattern didn't match.
    """

    match = self.match(regex, flags)
    if match:
      return match.group(group)
    return None

  def getline(self, cursor: Cursor) -> str:
    """ Returns the contents of the current line marked by the specified cursor location. """

    start = cursor.offset - cursor.column
    end = self.text.find('\n', start)
    return self.text[start:end]




__author__ = 'Niklas Rosenstein <rosensteinniklas@gmail.com>'
__version__ = '2.0.2'

from .scanner import Cursor, Scanner
from .tokenizer import RuleSet, Token, Tokenizer, TokenExtractor, TokenizationError, UnexpectedTokenError
from . import rules

__all__ = ['Cursor', 'Scanner', 'RuleSet', 'Token', 'Tokenizer', 'TokenExtractor',
  'TokenizationError', 'UnexpectedTokenError', 'rules']




import enum
import logging
import typing as t
from dataclasses import dataclass, field

from .ruleset import Rule, RuleConfigSet, RuleSet
from ..scanner import Cursor, Scanner

T = t.TypeVar('T')
U = t.TypeVar('U')
V = t.TypeVar('V')
R = t.TypeVar('R')


class Debug(enum.IntEnum):
  """ Flags that enable/disable various debug logs in the tokeninzer. """

  #: Log the final computed values for the "select" and "expect" arguments in the
  #: #Tokenizer.next() function.
  NEXT_ARGS = (1 << 0)

  #: Log the extracted token in #Tokenizer.next().
  EXTRACT = (1 << 1)

  #: Log when #Tokenizer.pos is set.
  UPDATE_POS = (1 << 2)

  NONE = 0
  ALL = (NEXT_ARGS | EXTRACT | UPDATE_POS)


@dataclass
class Token(t.Generic[T, U]):
  """
  Represents a token extracted by the tokenizer. When the end of the text is reached and no more
  tokens can be extracted, the *type* and *value* will be set to what is defined in the #RuleSet
  to be used as the sentinel.
  """

  type: T
  value: U
  pos: 'Cursor'
  eof: bool = field(repr=False)

  def __bool__(self) -> bool:
    return not self.eof

  @property
  def tv(self) -> t.Tuple[T, U]:
    return (self.type, self.value)


class ProxyToken(t.Generic[T, U]):
  """ Always represents the current token of the tokenizer. """

  def __init__(self,
    tokenizer: 'Tokenizer[T, U]',
    transformer: t.Optional[t.Callable[[Token[T, U]], Token[T, U]]] = None,
  ) -> None:
    self.tokenizer = tokenizer
    self.transformer = transformer
    self._transformed_token: t.Optional[t.Tuple[Token[T, U], Token[T, U]]] = None

  def __repr__(self) -> str:
    return f'TokenProxy({self._token!r})'

  def __bool__(self) -> bool:
    return bool(self._token)

  def __call__(self) -> Token[T, U]:
    return self._token

  @property
  def _token(self) -> Token[T, U]:
    if self.transformer is None:
      return self.tokenizer.current
    if self._transformed_token is not None and \
        self._transformed_token[0] is self.tokenizer.current:
      return self._transformed_token[1]
    self._transformed_token = (self.tokenizer.current, self.transformer(self.tokenizer.current))
    return self._transformed_token[1]

  @property
  def type(self) -> T:
    return self._token.type

  @property
  def value(self) -> U:
    return self._token.value

  @property
  def pos(self) -> Cursor:
    return self._token.pos

  @property
  def eof(self) -> bool:
    return self._token.eof

  @property
  def tv(self) -> t.Tuple[T, U]:
    return self._token.tv

  def next(self) -> None:
    self.tokenizer.next()

  def save(self) -> 'TokenizerState':
    return self.tokenizer.state

  def load(self, pos: 'TokenizerState') -> None:
    self.tokenizer.state = pos

  def set_ignored(self, token_types: t.Union[T, t.Collection[T]], ignored: bool = True) -> t.ContextManager[None]:
    return self.tokenizer.ignored.set(token_types, ignored)

  def set_skipped(self, token_types: t.Union[T, t.Collection[T]], skipped: bool = True) -> t.ContextManager[None]:
    return self.tokenizer.skipped.set(token_types, skipped)


@dataclass
class TokenizerState(t.Generic[T, U]):
  """ A checkpoint that can be used to restore the tokenizer to a previous state. """

  cursor: 'Cursor'
  token: t.Optional[Token[T, U]]
  skipped: RuleConfigSet[T, U, bool]
  ignored: RuleConfigSet[T, U, bool]
  skip_rule_once: t.Optional[Rule[T, U]]


class TokenizationError(Exception):
  """ Raised when the text cannot be tokenized. """


class UnexpectedTokenError(Exception):
  """ Raised when the text could only be tokenized into an unexpected token. """


class Tokenizer(t.Generic[T, U]):
  """ Splits text from a #Scanner into #Token#s via rules provided by a #Lexer. """

  log = logging.getLogger(__module__ + '.' + __qualname__)  # type: ignore

  #: The rule set that is used for tokenization.
  rules: 'RuleSet'

  #: The scanner that provides the text to be tokenized.
  scanner: 'Scanner'

  #: This config set defines the token types that are skipped when encountered (that means
  #: they are not returned by #next() or included when iterating over the tokenizer).
  skipped: RuleConfigSet[T, U, bool]

  #: This config set defines the token types that are ignored by the lexer (that means that
  #: are treated as if they would not exist in the rule set).
  ignored: RuleConfigSet[T, U, bool]

  def __init__(self, rules: 'RuleSet[T, U]', scanner: t.Union[str, 'Scanner'], debug: Debug = Debug.NONE) -> None:
    if isinstance(scanner, str):
      scanner = Scanner(scanner)
    self.rules = rules
    self.scanner = scanner
    self._current: t.Optional[Token[T, U]] = None
    self.skipped = RuleConfigSet(rules)
    self.ignored = RuleConfigSet(rules)
    self.debug = debug

    # Keep track if a zero-length token was extracted via a rule. That rule cannot trigger again
    # from the same position of the tokenizer.
    self._skip_rule_once: t.Optional[Rule[T, U]] = None

  def __bool__(self) -> bool:
    return bool(self.scanner) or bool(self._current)

  def __iter__(self) -> t.Iterator[Token[T, U]]:
    token = self.current
    while token:
      yield token
      token = self.next()

  @property
  def state(self) -> TokenizerState[T, U]:
    """ The position of the tokenizer. Can be set to go back to a previously stored position. """

    return TokenizerState(self.scanner.pos, self._current, self.skipped.copy(),
      self.ignored.copy(), self._skip_rule_once)

  @state.setter
  def state(self, state: TokenizerState[T, U]) -> None:
    if self.debug & Debug.UPDATE_POS:
      self.log.debug('Update Tokenizer.pos (pos=%r)', state)
    self.scanner.pos = state.cursor
    self._current = state.token
    self.skipped = state.skipped
    self.ignored = state.ignored
    self._skip_rule_once = state.skip_rule_once

  @property
  def current(self) -> Token[T, U]:
    """
    Returns the token that was last extracted in the #next() method. Raises an #EOF exception if
    the scanner has reached the end of the text. Raises a #RuntimeError if #next() has not been
    called successfully at least once. (You can check if the #initialized property returns `False`
    to see if the #RuntimeError would be raised).
    """

    if self._current is None:
      self.next()
      assert self._current is not None
    return self._current

  def next(self,
    select: t.Optional[t.Collection[T]] = None,
    unselect: t.Optional[t.Collection[T]] = None,
    expect: t.Optional[t.Collection[T]] = None,
  ) -> Token[T, U]:
    """
    Extracts the next token from the text and returns it. Without arguments, all currently
    non-ignored rules are considered. You may use the arguments to further reduce the set of
    rules used for parsing.

    If no token can be extracted using the specified rule subset, the tokenizer will attempt
    to tokenize using the remaining set of rules. If the extracted token is skippable, the next
    token will be extracted recursively. If the extracted token is not skippable, an
    #UnexpectedTokenError is raised.

    Note that the order in which rules are checked is always fixed to the order in which they
    occur on the rule set.

    # Arguments

    select (list): A list of token types that will be prioritized for extracting the next token.
      All token types passed into the *select* argument are taken into account as if they were
      also supplied to the *expect* argument.

    unselect (list): A list of token types that will be deprioritized for extracting the next
      token. The remaining token types are taken into account as if they were also supplied to the
      *expect* argument. The argument cannot be mixed with *select*.

    expect (list): A list of token types that are expected to be returned from this call. If the
      extracted token is not one of the specified token types, an #UnexpectedTokenError is raised.
      Skippable tokens are not skipped if they are expected via this argument.

    # Raises

    TokenizationError: If no token could be extracted from the current possition of the #scanner.

    UnexpectedTokenError: If the extracted token is not of one of the expected token types as
      specified via the *select*, *unselect* or *expect* arguments.

    ValueError: If one of the token types passed to *select*, *unselect* or *expect* is not a
      known token type.
    """

    if select is not None and unselect is not None:
      raise ValueError('`select` and `unselect` arguments cannot be mixed')

    if unselect is not None:
      unselect_set = set(unselect)
      self.rules.check_has_token_types(unselect_set)
      select = self.rules.token_types - unselect_set
    if select is not None:
      select_set = set(select)
      self.rules.check_has_token_types(select_set)
      expect = (t.cast(t.Set[T], set()) if expect is None else set(expect)) | select_set
    if expect is not None:
      if not isinstance(expect, set):
        expect = set(expect)
      self.rules.check_has_token_types(expect)

    if self.debug & Debug.NEXT_ARGS:
      self.log.debug('Tokenizer.next() computed arguments (select=%r, expect=%r)', select, expect)

    self._current = self._do_next(t.cast(t.Optional[t.Set[T]], select), expect)
    return self._current

  def _do_next(self, select: t.Optional[t.Set[T]], expect: t.Optional[t.Set[T]]) -> Token[T, U]:  # NOSONAR
    token, skippable = self._extract_token(lambda t: (select is None and not self.ignored.get(t, False)) or (select is not None and t in select))
    if token is None and select is not None:
      # Extract one of the unselected token types. If that token is skippable, we will accept
      # and skip it. If not, it will cause an #UnexpectedTokenError below given how we set up
      # the "expect" variable.
      token, skippable = self._extract_token(lambda t: select is not None and t not in select and not self.ignored.get(t, False))
      if token is not None and skippable:
        if self.debug & Debug.EXTRACT:
          self.log.debug('No selected token matched, but extract another skippable token "%s". '
            'Skip and continue recursively\n\t\ttoken: %r', token.type, token)
        return self._do_next(select, expect)
    elif token is not None and (expect is None or token.type not in expect) and skippable:
      if self.debug & Debug.EXTRACT:
        self.log.debug('Extracted skippable token "%s". Skip and continue recursively\n\t\ttoken: %r', token.type, token)
      return self._do_next(select, expect)

    if token is None:
      raise TokenizationError(self.scanner.pos)

    self.log.debug('Extracted token "%s"\n\t\ttoken: %r', token.type, token)

    if expect is not None and token.type not in expect:
      raise UnexpectedTokenError(token, expect)

    return token

  def _extract_token(self, filter: t.Callable[[T], bool]) -> t.Tuple[t.Optional[Token[T, U]], bool]:
    if not self.scanner:
      return Token(
        self.rules.sentinel.type,
        self.rules.sentinel.value,
        self.scanner.pos,
        True), False

    token_pos = self.scanner.pos
    for rule in self.rules:
      if not filter(rule.type) or rule == self._skip_rule_once:
        continue
      token_value = rule.extractor.get_token(self.scanner)
      if token_value is None:
        self.scanner.pos = token_pos
        continue
      token: Token[T, U] = Token(rule.type, token_value, token_pos, False)
      skippable = self.skipped.get(rule.type, rule.skip)
      if not token.value:
        # Zero-length token can only be produced once at a given location.
        # TODO(NiklasRosenstein): This only really works with strings as the token value.
        self._skip_rule_once = rule
      else:
        self._skip_rule_once = None
      return token, skippable
    return None, False

  Debug = Debug  # NOSONAR
  Error = TokenizationError
  Unexpected = UnexpectedTokenError




import re
import typing as t

from nr.parsing.core.tokenizer import TokenExtractor

if t.TYPE_CHECKING:
  from nr.parsing.core.scanner import Scanner


def regex(pattern: str, *, at_line_start_only: bool = False) -> TokenExtractor[re.Match]:
  """
  Creates a tokenizer rule that matches a regular expression and returns the #re.Match object
  as the token value. If you want the pattern to match at the start of a line only, set the
  *at_line_start_only* argument to `True`. The regex caret (`^`) control character will not work
  because the regex is matched from the cursor's current position and not the line start.
  """

  def _impl(scanner: 'Scanner') -> t.Optional[re.Match]:
    if at_line_start_only and scanner.pos.column != 0:
      return None
    match = scanner.match(pattern)
    if match is None:
      return None
    return match

  return TokenExtractor.of(_impl)


def regex_extract(pattern: str, group: t.Union[str, int] = 0, *,
    at_line_start_only: bool = False) -> TokenExtractor[str]:
  """
  Creates a token extractor that matches a regular expression and extracts a group from the match
  as the token value.
  """

  return regex(pattern, at_line_start_only=at_line_start_only).map(lambda m: m.group(group))


def string_literal(
  accepted_prefixes: t.Optional[str] = 'bfru',
  quote_sequences: t.Sequence[str] = ('"""', "'''", '"', "'"),
) -> TokenExtractor[str]:
  """
  Matches a Python string literal.
  """

  def _impl(scanner: 'Scanner') -> t.Optional[str]:
    prefix = (scanner.getmatch(r'[' + re.escape(accepted_prefixes) + r']+') or '') \
        if accepted_prefixes else ''
    quote_type = scanner.getmatch(r'(' + r'|'.join(re.escape(s) for s in quote_sequences) + r')')
    if not quote_type:
      return None
    is_multiline = len(quote_type) > 1
    contents = ''
    while scanner.char and (is_multiline or scanner.char != '\n'):
      if scanner.match(re.escape(quote_type)):
        break
      contents += scanner.char
      if scanner.char == '\\':
        contents += scanner.next()
      scanner.next()
    else:
      return None
    return prefix + quote_type + contents + quote_type

  return TokenExtractor.of(_impl)




import contextlib
import typing as t
from dataclasses import dataclass

from ..scanner import Scanner

if t.TYPE_CHECKING:
  from .extractor import TokenExtractor

T = t.TypeVar('T')
U = t.TypeVar('U')
V = t.TypeVar('V')
R = t.TypeVar('R')


@dataclass
class Sentinel(t.Generic[T, U]):
  type: T
  value: U


@dataclass
class Rule(t.Generic[T, U]):
  type: T
  extractor: 'TokenExtractor[U]'
  skip: bool


class RuleSet(t.Generic[T, U]):
  """
  A ordered list of parsing rules that is used a the #Tokenizer.
  """

  @t.overload
  def __init__(self: 'RuleSet[str, str]') -> None:
    """ The default constructor for a #RuleSet uses string for both the token and value type. """

  @t.overload
  def __init__(self: 'RuleSet[T, U]', sentinel: t.Union[t.Tuple[T, U], Sentinel[T, U]]) -> None:
    """ Initialize the #RuleSet with a sentinel configuration. """

  def __init__(self, sentinel = ('eof', '')) -> None:
    if isinstance(sentinel, tuple):
      sentinel = Sentinel(*sentinel)
    self._rules: t.List[Rule[T, U]] = []
    self._token_types: t.Set[T] = set()
    self.sentinel: Sentinel[T, U] = sentinel

  def __iter__(self) -> t.Iterator[Rule]:
    return iter(self._rules)

  @property
  def rules(self) -> t.List[Rule]:
    return list(self._rules)

  @property
  def token_types(self) -> t.Set[T]:
    return self._token_types

  def has_token_type(self, token_type: T) -> bool:
    return token_type in self._token_types

  def check_has_token_types(self, token_types: t.Set[T]) -> None:
    delta = token_types - self._token_types
    if delta:
      raise ValueError(f'unknown token types: {", ".join(map(str, delta))}')

  def rule(self, type_: T, extractor: 'TokenExtractor[U]', skip: bool = False) -> 'RuleSet[T, U]':
    """ Add a rule and return self. """

    self._rules.append(Rule(type_, extractor, skip))
    self._token_types.add(type_)
    return self


class RuleConfigSet(t.Generic[T, U, V]):
  """ Helper class to manage values associated with token types. """

  def __init__(self, rules: 'RuleSet[T, U]') -> None:
    self._rules = rules
    self._values: t.Dict[T, V] = {}

  def __repr__(self) -> str:
    return f'RuleConfigSet({self._values!r})'

  def set(self, token_types: t.Union[T, t.Collection[T]], value: V) -> t.ContextManager[None]:
    """
    Set the value of one or more token types. The returned context manager _may_ be used, but
    does not _have_ to be used, to revert to the previous state.

    Implementation detail: strings are not considered collections when identifying the type
    of the *token_types* argument.
    """

    if isinstance(token_types, str) or not isinstance(token_types, t.Collection):
      token_types_set = frozenset([t.cast(T, token_types)])
    else:
      token_types_set = frozenset(token_types)

    current_values = {k: v for k, v in self._values.items() if k in token_types_set}

    for token_type in token_types_set:
      if not self._rules.has_token_type(token_type):
        raise ValueError(f'not a possible token type: {token_type!r}')

    for token_type in token_types_set:
      self._values[token_type] = value

    @contextlib.contextmanager
    def _revert() -> t.Iterator[None]:
      try: yield
      finally:
        for token_type in token_types_set:
          if token_type not in current_values:
            # NOTE(NiklasRosenstein): https://github.com/python/mypy/issues/10152
            self._values.pop(token_type, None)  # type: ignore
          else:
            self._values[token_type] = current_values[token_type]

    return _revert()

  def get(self, token_type: T, default: R) -> t.Union[V, R]:
    return self._values.get(token_type, default)

  def copy(self) -> 'RuleConfigSet[T, U, V]':
    new = type(self)(self._rules)
    new._values.update(self._values)
    return new




# This file was auto-generated by Shut. DO NOT EDIT
# For more information about Shut, check out https://pypi.org/project/shut/

from __future__ import print_function
import io
import os
import setuptools
import sys

def _tempcopy(src, dst):
  import atexit, shutil
  if not os.path.isfile(dst):
    if not os.path.isfile(src):
      print('warning: source file "{}" for destination "{}" does not exist'.format(src, dst))
      return
    shutil.copyfile(src, dst)
    atexit.register(lambda: os.remove(dst))


_tempcopy('../LICENSE.txt', 'LICENSE.txt')

readme_file = 'README.md'
if os.path.isfile(readme_file):
  with io.open(readme_file, encoding='utf8') as fp:
    long_description = fp.read()
else:
  print("warning: file \"{}\" does not exist.".format(readme_file), file=sys.stderr)
  long_description = None

requirements = []
test_requirements = [
  'pytest',
]

setuptools.setup(
  name = 'nr.parsing.core',
  version = '2.0.2',
  author = 'Niklas Rosenstein',
  author_email = 'rosensteinniklas@gmail.com',
  description = 'A simple API to scan and tokenize text for the purpose of structured language processing.',
  long_description = long_description,
  long_description_content_type = 'text/markdown',
  url = 'https://git.niklasrosenstein.com/NiklasRosenstein/nr',
  license = 'MIT',
  packages = setuptools.find_packages('src', ['test', 'test.*', 'tests', 'tests.*', 'docs', 'docs.*']),
  package_dir = {'': 'src'},
  include_package_data = True,
  install_requires = requirements,
  extras_require = {},
  tests_require = test_requirements,
  python_requires = '>=3.6.0,<4.0.0',
  data_files = [],
  entry_points = {},
  cmdclass = {},
  keywords = [],
  classifiers = [],
  zip_safe = False,
)
